/******************************************************************************
 * Copyright (c) 2014, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *   ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *****************************************************************************/

/**
 * \file data_slice.h
 * \brief Abstraction of a single slice of data used by an Accelerator 
 */

#ifndef DATA_SLICE_H_
#define DATA_SLICE_H_

#define __CL_ENABLE_EXCEPTIONS
#include <CL/cl.hpp>
#include "ocl_util.h" 
#include <iostream>
#include <iomanip>
#include <cstdint>
#include <utility>
#include <string>
#include <functional>
#include "types.h"

#include "data_slice_id.h"

/**
 * \class DataSlice
 * \brief Representation of a host data section (array slice) 
 *
 * A data section on the host with a corresponding CL device buffer 
 * that can be used to map host and device memory and make related
 * data transfers. The CL device buffer may not always be initialized.
 * The value of init_ is only true when a DataSlice has an initialized 
 * DataSliceID id_ (i.e. has a start and end address).
 */
class DataSlice
{

    public:
        /**
         * \brief Default Constructor
         */
        DataSlice() { init_ = false; }
        
        /**
         * \brief Constructor
         * \param size  size (in bytes) of data slice
         * \param kind  MapKind of DataSlice
         * \param debug Enable/Disable debugging information
         * \param perf  Enable/Disable performance information
         */
        DataSlice(unsigned int size, 
                  MapKind kind, 
                  bool debug = false,
                  bool perf = false
                  );
        
        /**
         * \brief Constructor
         * \param start Start address of data slice
         * \param size  size (in bytes) of data slice
         * \param kind  MapKind of DataSlice
         * \param debug Enable/Disable debugging information
         * \param perf  Enable/Disable performance information
         */
        DataSlice(void* start, 
                  unsigned int size, 
                  MapKind kind, 
                  bool debug = false,
                  bool perf = false
                  );
       
        /**
         * \brief Destructor
         */
        ~DataSlice() {}
       
        /**
         * \brief Return start address of slice
         * \return Start address of slice
         */
        void* start() const 
        { return reinterpret_cast<void*>(id_.start()); }
       
         /**
         * \brief Set start address of slice
         * \param start Pointer to start memory location
         */
        void set_start(void* start);

        /**
         * \brief Return integer value of parent slice's start address
         * \return Integer value of parent slice start address
         */
        DataSliceID parent_id() const { return parent_id_; }
        
        /**
         * \brief Set value of parent slice's start address
         */
        void set_parent_id(const DataSliceID& parent_id) 
        { parent_id_ = parent_id; }

        /**
         * \brief Return start address of slice
         * \return Start address of slice
         */
        void* end() const 
        { return reinterpret_cast<void*>(id_.end()); }
 
        /**
         * \brief Return size of slice
         * \return Size (in bytes) of slice
         */
        unsigned int size() const { return size_; }
 
        /**
         * \brief Return kind of DataSlice
         * \return MapKind of DataSlice
         */
        MapKind kind() const { return kind_; }
 
        /**
         * \brief Set the kind of DataSlice
         */
        void set_kind(MapKind kind) { kind_ = kind; }
 
        /**
         * \brief Return reference to OpenCL device buffer
         * \return Reference to CL Buffer
         */
        cl::Buffer& device_buffer() { return device_buffer_; }
        
        /**
         * \brief Return reference to OpenCL event
         * \return Reference to CL Event
         */
        cl::Event& event() { return event_; }

        /**
         * \brief Return unique ID
         * \return Reference to DataSliceID
         */
        const DataSliceID& id() const { return id_; }

        /**
         * \brief Check if a data slice contains a memory address
         * \return true if the address falls within the slice borders
         */
        bool Contains(void* address) const 
        { return id_.Contains(reinterpret_cast<uintptr_t>(address)); }

        /**
         * \brief Check if a data slice contains given data slice
         * \return true if the given slice falls completely within the 
         *         slice borders
         */
        bool Contains(const DataSlice& slice) const
        { return id_.Contains(slice.id()); }

        /**
         * \brief Make this slice a subslice of the given 
         *        parent DataSlice i.e. create a CL sub-buffer out 
         *        of parent's CL buffer             
         */
        void MakeSubSliceOf(DataSlice& parent_slice);
        
        /**
         * \brief Initialize the CL device buffer for this slice 
         */
        void InitDeviceBuffer(cl::Context& device_context);

        /**
         * \brief Get OpenCL memory flags associated with a MapKind
         * \return cl_mem_flags value
         */
        cl_mem_flags GetCLMemFlags(MapKind kind);

        /**
         * \brief Initiate transfer of host slice data to device buffer 
         * \param queue     CL command queue to enque write on
         * \param blocking  True if synchronous write, false if
         *                  asynchronous write needs to be initiated
         */
        void WriteToDevice(cl::CommandQueue& queue, bool blocking);
        
        /**
         * \brief Initiate transfer of device buffer data to host 
         * \param queue     CL command queue to enque write on
         * \param blocking  True if synchronous read, false if
         *                  asynchronous read needs to be initiated
         */
        void ReadFromDevice(cl::CommandQueue& queue, bool blocking);
    
        /**
         * \brief Wait for a write event to complete 
         */
        void WaitOnWrite();
     
        /**
         * \brief Wait for a read event to complete 
         */
        void WaitOnRead();   
    
        /**
         * \brief Get time elapsed for a read/write on this slice
         *        in microseconds
         */
        cl_ulong GetExecutionTime() const;

        /**
         * \brief Show read/write performance data
         * \param type Description of type of event that is being 
         *             shown i.e. Host-> Device or Device->Host
         */
        void ShowPerformanceData(std::string type);

        /**
         * \brief Make a write to device buffer or read data from 
         *        device buffer based on given kind 
         * \param queue    CL command queue to enque write on
         * \param kind     MAP_TO or MAP_FROM only
         */
        void Update(cl::CommandQueue& queue, MapKind kind);
        
        /**
         * \brief Compare two data slices
         * \return true if both slices have same start and size
         */
        bool operator==(const DataSlice& slice) const;
 
    private:
        bool            init_;
        bool            debug_;
        bool            show_perf_;
        unsigned int    size_;
        MapKind         kind_;
        DataSliceID     id_;
        DataSliceID     parent_id_;
        cl::Event       event_;
        cl::Buffer      device_buffer_;
};


/**
 * \brief Pretty print DataSlice
 * \param os    Output stream reference 
 * \param slice DataSlice reference
 * \return Reference to output stream object  
 */
inline std::ostream& operator<<(std::ostream& os, const DataSlice& slice)
{
    const int field_width = 10;
    const int address_width = 12;
    const int size_width = 12;
    const int kind_width = 18;

    os  << std::left  << std::setw(field_width)   << "> [start: "  
        << std::right << std::setw(address_width) << slice.start() << ", "
        << std::left  << std::setw(field_width)   << "end: "  
        << std::right << std::setw(address_width) << slice.end() << ", "
        << std::left  << std::setw(field_width)   << "size: " 
        << std::right << std::setw(size_width)    << slice.size() << "b, "
        << std::left  << std::setw(field_width)   << "map-type: " 
        << std::right << std::setw(kind_width)    << slice.kind()
        << "]";

    return os;
}

#endif /* DATA_SLICE_H_ */
