/******************************************************************************
 * Copyright (c) 2014, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *   ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *****************************************************************************/

/**
 * \file ompacc.cc
 * \brief OpenMP 4.0 Accelerator Model API 
 */

#include "ompacc.h"

/*********************************************************************/
/* Single static instance of Accelerator object                      */
/*********************************************************************/
static Accelerator dsp;

/*********************************************************************/
/* Setup data items specific to a target data region                 */
/*********************************************************************/
void GOMP_target_data (  int device,            
        unsigned int mapnum,   
        void **hostaddrs,     
        unsigned int *sizes,   
        unsigned char *kinds   
        )
{
    dsp.InitializeCL(device);
    dsp.SetupTargetDataRegion(mapnum, hostaddrs, sizes, kinds);
}

/*********************************************************************/
/* Destroy data for a target data region                             */
/*********************************************************************/
void GOMP_target_data_end (  int device,            
        unsigned int mapnum,   
        void **hostaddrs,     
        unsigned int *sizes,   
        unsigned char *kinds   
        )
{
    dsp.DestroyTargetDataRegion(mapnum, hostaddrs, sizes, kinds);
}

/*********************************************************************/
/* Make a target update for given data items                         */
/*********************************************************************/
void GOMP_target_update (  int device,            
        unsigned int mapnum,   
        void **hostaddrs,     
        unsigned int *sizes,   
        unsigned char *kinds   
        )
{
   dsp.MakeTargetUpdate(mapnum, hostaddrs, sizes, kinds); 
}


/*********************************************************************/
/* Malloc a data region in the given memory segment with given size  */
/* and return a pointer mapped to it                                 */
/*********************************************************************/
void* __TI_omp_device_alloc(int device_id, 
                            __TI_omp_mem_region mregion, 
                            int size
                            )
{
    dsp.InitializeCL(device_id);
    return dsp.CreateMapBuffer(mregion, 
                               size
                               ); 
} 

/*********************************************************************/
/* Free a data region which has the given pointer mapped to it       */
/*********************************************************************/
void __TI_omp_device_free(int device_id, void* buffer)
{
    /* NOTE: device_id not used at the moment but is passed in to 
     *       support cases in future when this runtime will offload 
     *       to multiple accelerator devices 
     *       */
    dsp.FreeMapBuffer(buffer);    
}

/*********************************************************************/
/* Create a target region on the given device and execute the given  */
/* function on it                                                    */
/*********************************************************************/
void GOMP_target (  int device, 
        void (*fn) (void *),  
        char *fnname,  
        unsigned int mapnum,      
        void **hostaddrs,  
        unsigned int *sizes,    
        unsigned char *kinds
        )
{
    dsp.InitializeCL(device);
   
    dsp.InitializeTargetRegion(mapnum, hostaddrs, sizes, kinds);

    try 
    {
        /*********************************************************************/
        /* Initialize buffers for kernel arguments                           */
        /*********************************************************************/
        dsp.CreateKernelArgs(); 
   
        /*********************************************************************/
        /* Create kernel and set kernel arguments                            */        
        /*********************************************************************/
        std::string k_fn(fnname);
        k_fn.append("_wrapper");
        dsp.BuildKernelAndSetArgs(k_fn);
        
        /*********************************************************************/
        /* Enqueue write buffers - blocking                                  */
        /*********************************************************************/
        dsp.EnqueueWriteBuffers();
        
        /*********************************************************************/
        /* Enqueue kernel                                                    */
        /*********************************************************************/
        dsp.EnqueueKernel();
        
        /*********************************************************************/
        /* Show performance data for kernel execution                        */
        /*********************************************************************/
        dsp.ReportKernelPerformance();

        /*********************************************************************/
        /* Enqueue read buffers - blocking                                   */
        /*********************************************************************/
        dsp.EnqueueReadBuffers();
        
        /*********************************************************************/
        /* Invalidate Map Buffers                                            */
        /*********************************************************************/
        dsp.InvAllMapBuffers(kinds);

        /*********************************************************************/
        /* Make ready for next call to GOMP_target */
        /*********************************************************************/
        dsp.ClearTargetData();

    }
    catch (Error err) 
    { 
        std::cerr << ">> ERROR: GOMP_Target: " << err.what() 
                  << "(" << err.err() << ")" << std::endl; 
    }
}
